<?php 
/**
 * Backup script for Icebear.
 * 
 * This backup script will do NOTHING unless Icebear is configured with both a backup storage area
 * and a number of backups to keep. Obviously the backup area should be on a different disk, ideally 
 * in a different machine in a different building (or town, or country, depending on your paranoia level).
 * 
 * Rationale: 
 * 
 * Imported images will be written to both the live store and the backup store as they
 * are generated. Therefore, this backup script only needs to take care of the database and the Icebear 
 * application itself.
 * 
 * The directory containing uploaded files will be rsynced to its backup directory, if the backup is specified. 
 * 
 * The database will be effectively read-only for the duration of its backup, so we write a small file
 * before starting the database backup and remove it afterward. Thus, imager importers can check for
 * the existence of this file and abort if found.
 * 
 */


//If called from web, bail - command line/cron only. .htaccess should do this but be safe.
if('cli'!==php_sapi_name()){
	die("Backup script can only be called from command line or cron");
}

define('LOGLEVEL_DEBUG', 1);
define('LOGLEVEL_INFO', 2);
define('LOGLEVEL_WARN', 3);
define('LOGLEVEL_ERROR', 4);
$logLevel=LOGLEVEL_INFO; //may be overridden by argument

set_time_limit(0);
setUpClassAutoLoader();

for($i=1;$i<count($argv);$i++){
	$arg=$argv[$i];
	if(preg_match('/^-h$/',$arg)){
		showHelp();
		exit();
	} else if(preg_match('/^-l[1-4]$/',$arg)){
		$logLevel=(int)substr($arg, 2);
	}
}

writeLog(LOGLEVEL_INFO,'Beginning backup of Icebear database and www directory');

try{
	//get db connection details from conf/config.ini
	$dbDetails=parse_ini_file(realpath(__DIR__).'/../conf/config.ini');
	if(!$dbDetails){ 
		throw new Exception('Could not read database connection details from config file'); 
	}
	$db = new PDO('mysql:host='.$dbDetails['dbHost'].';dbname='.$dbDetails['dbName'], $dbDetails['dbUser'], $dbDetails['dbPass'], array(PDO::MYSQL_ATTR_FOUND_ROWS => true));
	$db->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);


	//connect to database
	writeLog(LOGLEVEL_DEBUG,'Attempting to get LIMS DB connection... ');
	database::connect();
	writeLog(LOGLEVEL_DEBUG,'...got LIMS DB connection');

	writeLog(LOGLEVEL_DEBUG,'Attempting to get LIMS session... ');
	session::init(new DummySession());
	session::set('isAdmin',true);
	writeLog(LOGLEVEL_DEBUG,'...got LIMS session');

	//get config items
	$backupStorage=config::get('core_databasebackup');
	$backupsToKeep=config::get('core_databasebackupstokeep');
	$backupsToKeep=(int)$backupsToKeep;
	$backupsToKeep=max($backupsToKeep,0);
	
	//destroy LIMS session, no longer needed
	session::destroy();
	
	if(0==$backupsToKeep || empty(rtrim($backupStorage,'/\\'))){
		writeLog(LOGLEVEL_WARN, 'Not backing up because configuration says not to.');
		writeLog(LOGLEVEL_WARN, 'Any existing backups will remain in place.');
		throw new Exception('Config item "Backups to keep" set to zero or backups directory not defined, not backing up database');
	}
	
	//verify that core_icebearbackups exists and can be written
	if(!@file_exists($backupStorage)){
		throw new Exception("Directory for database backups does not exist: ".$backupStorage);
	}
	
	//touch .backupinprogress - importer should check for existence and bail if found
	$backupStorage=rtrim($backupStorage,'/\\').'/';
	writeLog(LOGLEVEL_DEBUG, 'Writing lock file before database backup...');
	if(!@touch($backupStorage.'dbbackuprunning')){
		throw new Exception("Could not create lock file before database backup");
	}
	writeLog(LOGLEVEL_DEBUG, '...Done.');
	
	//dump the database to .sql file in backup storage root
        writeLog(LOGLEVEL_DEBUG, 'Commencing database dump...');
	$sqlFilename='icebear.sql';
	$cmd='mysqldump '.$dbDetails['dbName'].' --password=\''.$dbDetails['dbPass'].'\' --user=\''.$dbDetails['dbUser'].'\' --single-transaction >'.$backupStorage.$sqlFilename;
	$result=exec($cmd,$output);
        writeLog(LOGLEVEL_DEBUG, '...Database dump done.');
	
	//remove lock file
	writeLog(LOGLEVEL_DEBUG, 'Removing lock file after database backup...');
	if(!@unlink($backupStorage.'dbbackuprunning')){
		throw new Exception("Could not remove lock file after database backup");
	}
	writeLog(LOGLEVEL_DEBUG, '...Done.');

	//Bail if database dump failed
	if(!empty($output)){
		throw new Exception($output);
	}
	
	//make .tar.gz from .sql and /var/www - assume backup script is located at [wwwroot]/backup/backup.php 
	$parts=explode('/backup', __DIR__);
	$wwwroot=$parts[0];
	$result=system('tar -czf '.$backupStorage.'icebear_inprogress.tar.gz '.$backupStorage.$sqlFilename.' '.$wwwroot);
	
	//remove the .sql file
	@unlink($backupStorage.$sqlFilename);
	
	//rotate older backups on core_icebearbackups
	//Note that we DO NOT remove older backups beyond the specified period. A malicious Icebear admin could have reduced the
	//backup period before causing mischief.
	for($i=$backupsToKeep-1;$i>0;$i--){
		@rename($backupStorage.'icebear.tar.gz.'.$i, $backupStorage.'icebear.tar.gz.'.($i+1));
	}
	@rename($backupStorage.'icebear.tar.gz', $backupStorage.'icebear.tar.gz.1');
	
	//Rename new backup into place
	@rename($backupStorage.'icebear_inprogress.tar.gz', $backupStorage.'icebear.tar.gz');
	
	
	
	writeLog(LOGLEVEL_DEBUG, 'File uploads: Synching to backup...');
	$filesBackup=config::get('core_filestorebackup');
	if(empty($filesBackup)){
		writeLog(LOGLEVEL_INFO, 'core_filestorebackup not set - not backing up file uploads');
	}else {
		$filesStore=config::get('core_filestore');
		$filesStore=rtrim($filesStore,'/').'/';
		$filesBackup=rtrim($filesBackup,'/').'/';
		$cmd='rsync -qa '.$filesStore.' '.$filesBackup;
		writeLog(LOGLEVEL_DEBUG, 'Source: '.$filesStore);
		writeLog(LOGLEVEL_DEBUG, 'Dest: '.$filesBackup);
		writeLog(LOGLEVEL_DEBUG, 'Command: '.$cmd);
		$output='';
		exec($cmd,$output);
		if(!empty($output)){
			throw new Exception($output);
		}
	}
	writeLog(LOGLEVEL_DEBUG, '...Done');
	
	

} catch(Exception $e){
	writeLog(LOGLEVEL_ERROR, get_class($e).' caught: '.$e->getMessage());
	$trace=$e->getTrace();
	foreach($trace as $t){
		writeLog(LOGLEVEL_ERROR, ': '.$t['file']);
		if(isset($t['type']) && isset($t['class'])){
			writeLog(LOGLEVEL_ERROR, ':    Line '.$t['line'].', Function '.$t['class'].$t['type'].$t['function']);
		} else {
			writeLog(LOGLEVEL_ERROR, ':    Line '.$t['line'].', Function '.$t['function']);
		}
		//TODO $t['args']
	}
	writeLog(LOGLEVEL_ERROR, 'Backup threw exception. Ending.');
}

writeLog(LOGLEVEL_INFO,'Backup script finished.');
writeLog(LOGLEVEL_INFO,'=================================================');


exit; //functions below

/**
 * Echoes a log message to the output, prefixed by a level indicator and date/time.
 * @param int $level The log level, see top of this file for definitions
 * @param string $message The log message
 */

function writeLog($level, $message){
	global $logLevel;
	if($level<$logLevel){ return; }
	$labels=array(
			LOGLEVEL_DEBUG=>'DEBUG',
			LOGLEVEL_INFO =>'INFO ',
			LOGLEVEL_WARN =>'WARN ',
			LOGLEVEL_ERROR=>'ERROR',
	);
	$out=$labels[$level].' '.date('d M y H:i:s').' '.$message."\n";
	echo $out;
}

function setUpClassAutoLoader(){
	spl_autoload_register(function($className){
		$paths=array(
				'../classes/',
				'../classes/core/',
				'../classes/core/exception/',
				'../classes/core/authentication/',
				'../classes/core/interface/',
				'../classes/model/',
		);
		foreach($paths as $path){
			if(file_exists(__DIR__.'/'.$path.$className.'.class.php')){
				include_once(__DIR__.'/'.$path.$className.'.class.php');
			}
		}
	});
}

function showHelp(){
	echo "Backs up the Icebear database and the web application files\n";
	echo "Call with -l1 for full debug info, 2 is info (default), 3 only warnings and errors, 4 errors only\n";
}

?>
